unit ScrollBar_Plugin;

interface


uses Windows, ComObj, ActiveX, SysUtils, SCROLLBARS_TLB, AQTest_TLB, Messages;

const
  WM_CALL = WM_USER + 1;


type
  TScrollCallbackProc = procedure(nCode:Integer; wParam:WParam; lParam:LPARAM); stdcall;
  TScrollHookProc = procedure; stdcall;

  TGetHookRecPointer = function : pointer stdcall;

  PHookRec = ^THookRec;
  THookRec = packed record
    TheCallBack : HWND;
    Msg:TMsg;
  end;

  TCurrentScrollBar = (csbNone, csbHorz, csbVert);

  TScrollBarPlugIn = class(TInterfacedObject, IPTEngine, IPTSettingsNotifier, IPTEngineNotifier, IPTLowRecorderNotifier)
  private
    FPTSettings: IPTSettings;
    FScriptLang:TScriptLang;
    FIcon: HICON;
    FEnabled: Boolean;
    FHandle:THandle;
    FHWND:HWND;
    FStartHook:TScrollHookProc;
    FStopHook:TScrollHookProc;
    FHookRec:PHookRec;
    FGetHookRec:TGetHookRecPointer;
    FHScrollPOS:Integer;
    FVScrollPOS:Integer;
    FHSaveScrollPOS:Integer;
    FVSaveScrollPOS:Integer;
    FScrolling: Boolean;
    FSavedObj:OleVariant;
    procedure OnHookWMCommand(wNotifyCode, wID: Word; hwndCtl: Integer);
  public
    //name of the tool
    function Get_Name: WideString; safecall;
    property Name: WideString read Get_Name;
    //category of the tool
    function Get_Category: WideString; safecall;
    property Category: WideString read Get_Category;
    //icon
    function Get_Icon: SYSUINT; safecall;
    property Icon: SYSUINT read Get_Icon;
    //interface to the settings that can be configured by the end-user
    function Get_Settings: IPTSettings; safecall;
    property Settings: IPTSettings read Get_Settings;
    property Enabled: Boolean read FEnabled write FEnabled;
    // IPTEngineNotifier
    procedure OnRecordStart; safecall;
    procedure OnRecordStop; safecall;
    procedure OnRecordPause; safecall;
    procedure OnRecordResume; safecall;
    procedure OnExecuteStart(const ProcedureName: WideString); safecall;
    procedure OnExecuteStop; safecall;
    procedure OnExecutePause; safecall;
    procedure OnExecuteResume; safecall;
    procedure OnChangeState(State: TEngineState); safecall;
    // IPTSettingsNotifier
    procedure UpdateSettings; safecall;
    //IPTLowRecorderNotifier
    procedure OnKeyDown(Scan: Integer; VK: Integer); safecall;
    procedure OnKeyUp(Scan: Integer; VK: Integer); safecall;
    procedure OnMouseDown(VK: Integer; X: Integer; Y: Integer); safecall;
    procedure OnMouseUp(VK: Integer; X: Integer; Y: Integer); safecall;
    procedure OnMouseMove(X: Integer; Y: Integer); safecall;
    procedure OnMouseWheel(zDelta: Integer); safecall;

    procedure AddScrollLine(Name:String);
    function GetCurrentScrollPos: Boolean;

    constructor Create;
    destructor Destroy; override;
    procedure WinProc(var Message: TMessage);
  end;

var
  SCROLLBARPlugIn: TSCROLLBARPlugIn;
  PIManager: IPIManager;

{ export functions }
function AQtestPlugInInit(PlugInManager: IPIBaseManager): WordBool; stdcall;
procedure AQtestPlugInFinalize; stdcall;
function AQtestPlugInDescription: WideString; stdcall;


implementation


uses SCrollBarImpl
{$IFDEF Codesite}
,CSIntf
{$ENDIF}
;

type
  TWndMethod = procedure(var Message: TMessage) of object;

const
  InstanceCount = 313;


type
  PObjectInstance = ^TObjectInstance;
  TObjectInstance = packed record
    Code: Byte;
    Offset: Integer;
    case Integer of
      0: (Next: PObjectInstance);
      1: (Method: TWndMethod);
  end;

type
  PInstanceBlock = ^TInstanceBlock;
  TInstanceBlock = packed record
    Next: PInstanceBlock;
    Code: array[1..2] of Byte;
    WndProcPtr: Pointer;
    Instances: array[0..InstanceCount] of TObjectInstance;
  end;

var
  InstBlockList: PInstanceBlock;
  InstFreeList: PObjectInstance;


{ Standard window procedure }
{ In    ECX = Address of method pointer }
{ Out   EAX = Result }

function StdWndProc(Window: HWND; Message, WParam: Longint;
  LParam: Longint): Longint; stdcall; assembler;
asm
        XOR     EAX,EAX
        PUSH    EAX
        PUSH    LParam
        PUSH    WParam
        PUSH    Message
        MOV     EDX,ESP
        MOV     EAX,[ECX].Longint[4]
        CALL    [ECX].Pointer
        ADD     ESP,12
        POP     EAX
end;



{ Allocate an object instance }

function CalcJmpOffset(Src, Dest: Pointer): Longint;
begin
  Result := Longint(Dest) - (Longint(Src) + 5);
end;

function MakeObjectInstance(Method: TWndMethod): Pointer;
const
  BlockCode: array[1..2] of Byte = (
    $59,       { POP ECX }
    $E9);      { JMP StdWndProc }
  PageSize = 4096;
var
  Block: PInstanceBlock;
  Instance: PObjectInstance;
begin
  if InstFreeList = nil then
  begin
    Block := VirtualAlloc(nil, PageSize, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
    Block^.Next := InstBlockList;
    Move(BlockCode, Block^.Code, SizeOf(BlockCode));
    Block^.WndProcPtr := Pointer(CalcJmpOffset(@Block^.Code[2], @StdWndProc));
    Instance := @Block^.Instances;
    repeat
      Instance^.Code := $E8;  { CALL NEAR PTR Offset }
      Instance^.Offset := CalcJmpOffset(Instance, @Block^.Code);
      Instance^.Next := InstFreeList;
      InstFreeList := Instance;
      Inc(Longint(Instance), SizeOf(TObjectInstance));
    until Longint(Instance) - Longint(Block) >= SizeOf(TInstanceBlock);
    InstBlockList := Block;
  end;
  Result := InstFreeList;
  Instance := InstFreeList;
  InstFreeList := Instance^.Next;
  Instance^.Method := Method;
end;

{ Free an object instance }

procedure FreeObjectInstance(ObjectInstance: Pointer);
begin
  if ObjectInstance <> nil then
  begin
    PObjectInstance(ObjectInstance)^.Next := InstFreeList;
    InstFreeList := ObjectInstance;
  end;
end;

var
  UtilWindowClass: TWndClass = (
    style: 0;
    lpfnWndProc: @DefWindowProc;
    cbClsExtra: 0;
    cbWndExtra: 0;
    hInstance: 0;
    hIcon: 0;
    hCursor: 0;
    hbrBackground: 0;
    lpszMenuName: nil;
    lpszClassName: 'TPUtilWindow');

function AllocateHWnd(Method: TWndMethod): HWND;
var
  TempClass: TWndClass;
  ClassRegistered: Boolean;
begin
  UtilWindowClass.hInstance := HInstance;
  ClassRegistered := GetClassInfo(HInstance, UtilWindowClass.lpszClassName,
    TempClass);
  if not ClassRegistered or (TempClass.lpfnWndProc <> @DefWindowProc) then
  begin
    if ClassRegistered then
      Windows.UnregisterClass(UtilWindowClass.lpszClassName, HInstance);
    Windows.RegisterClass(UtilWindowClass);
  end;
  Result := CreateWindowEx(WS_EX_TOOLWINDOW, UtilWindowClass.lpszClassName,
    '', WS_POPUP {!0}, 0, 0, 0, 0, 0, 0, HInstance, nil);
  if Assigned(Method) then
    SetWindowLong(Result, GWL_WNDPROC, Longint(MakeObjectInstance(Method)));
end;

procedure DeallocateHWnd(Wnd: HWND);
var
  Instance: Pointer;
begin
  Instance := Pointer(GetWindowLong(Wnd, GWL_WNDPROC));
  DestroyWindow(Wnd);
  if Instance <> @DefWindowProc then FreeObjectInstance(Instance);
end;





procedure TScrollBarPlugIn.AddScrollLine(Name:String);
var
  s:string;
begin
   if GetCurrentScrollPos and ((FHScrollPOS <> FHSaveScrollPOS ) or
     (FVScrollPOS <> FVSaveScrollPOS))  then
    with PIManager.Recorder do
    begin
      s := Format('  ScrollBars.ScrollBy(w.%s.ID,%d,%d)',[Name,FHScrollPOS,FVScrollPOS]);
      if FScriptLang = DelphiScript then
        s := s + ';';

      AddString(s);

      FHSaveScrollPOS := FHScrollPOS;
      FVSaveScrollPOS := FVScrollPOS;
    end;
end;



function LoadTypeLibrary(const ModuleName: string): ITypeLib;
var
  wVerMajor:Word;
  wVerMinor:Word;
  aLCID:lcid;
begin
  OleCheck(LoadRegTypeLib(LIBID_AQtest,wVerMajor,wVerMinor, aLCID, Result));
end;

constructor TScrollBarPlugIn.Create;
begin
//  inherited Create(LoadTypeLibrary('..\AQTest.Exe'),IPTEngine);
  inherited Create;
  if FSCROLLBARS = nil then
    FSCROLLBARS := TScrollBar.Create() as IDispatch;
  FHandle := LoadLibrary('AQScrollHook.Dll');
  if  FHandle <> 0 then
  begin
    @FStartHook := GetProcAddress(FHandle, 'StartHook');
    @FStopHook := GetProcAddress(FHandle, 'StopHook');
    @FGetHookRec := GetProcAddress(FHandle, 'GetHookRecPointer');
  end;
  FHWND := AllocateHWnd(WinProc);
end;

destructor TScrollBarPlugIn.Destroy;
begin
  DeallocateHWnd(FHWND);
  FreeLibrary(FHandle);
  inherited;
end;


function TScrollBarPlugIn.GetCurrentScrollPos: Boolean;
var
 Err:DWord;
 si: TScrollInfo;
begin

//  FHScrollPOS := GetScrollPos(FSavedObj.ID,SB_HORZ);
//  FVScrollPOS := GetScrollPos(FSavedObj.ID,SB_VERT);
//
//  Codesite.SendInteger('Horz',FHScrollPOS);
//  Codesite.SendInteger('Vert',FVScrollPOS);

   FillChar( si, sizeof( si ), 0 ); // always a good idea
   with si do begin
     cbSize := sizeof(si);
     fMask := SIF_ALL;
   end;
   GetScrollInfo( FSavedObj.ID, SB_HORZ, si );

   FHScrollPOS := si.nPos;



   FillChar( si, sizeof( si ), 0 ); // always a good idea
   with si do begin
     cbSize := sizeof(si);
     fMask := SIF_ALL;
   end;
   GetScrollInfo( FSavedObj.ID, SB_VERT, si );

   FVScrollPOS := si.nPos;


end;

function TScrollBarPlugIn.Get_Category: WideString;
begin
  Result := 'Script';
end;


function TScrollBarPlugIn.Get_Icon: SYSUINT;
begin
  Result := FIcon;
end;

function TScrollBarPlugIn.Get_Name: WideString;
begin
  Result := 'Scrollbar support';
end;

function TScrollBarPlugIn.Get_Settings: IPTSettings;
begin
  Result := FPTSettings;
end;

procedure TScrollBarPlugIn.OnChangeState(State: TEngineState);
begin

end;

procedure TScrollBarPlugIn.OnExecutePause;
begin

end;

procedure TScrollBarPlugIn.OnExecuteResume;
begin

end;

procedure TScrollBarPlugIn.OnExecuteStart(const ProcedureName: WideString);
begin
end;

procedure TScrollBarPlugIn.OnExecuteStop;
begin

end;


procedure TScrollBarPlugIn.OnHookWMCommand(wNotifyCode, wID: Word;
  hwndCtl: Integer);
begin
{$IFDEF Codesite}
CodeSite.SendInteger('wNotifyCode', wNotifyCode);
CodeSite.SendInteger('wID', wID);
Codesite.SendWindowHandle('hwndCtl',hwndCtl);
{$ENDIF}
end;

procedure TScrollBarPlugIn.OnKeyDown(Scan, VK: Integer);
begin
  if FScrolling and FEnabled then
    AddScrollLine(FSavedObj.Name);
end;

procedure TScrollBarPlugIn.OnKeyUp(Scan, VK: Integer);
begin
end;

procedure TScrollBarPlugIn.OnMouseDown(VK, X, Y: Integer);
begin
  if FEnabled then
  begin
    FSavedObj := PIManager.SYS.ObjectFromPoint(X,Y);
    {$IFDEF Codesite}
    CodeSite.SendString('FSavedObj.Name', String(FSavedObj.Name));
    CodeSite.SendWindowHandle('FSavedObj.ID',FSavedObj.ID);
    {$ENDIF}
  end;
end;

procedure TScrollBarPlugIn.OnMouseMove(X, Y: Integer);
begin

end;

procedure TScrollBarPlugIn.OnMouseUp(VK, X, Y: Integer);
begin
  if FScrolling and FEnabled then
  begin
    FSavedObj := PIManager.SYS.ObjectFromPoint(X,Y);
    {$IFDEF Codesite}
    CodeSite.SendString('FSavedObj.Name', String(FSavedObj.Name));
    CodeSite.SendWindowHandle('FSavedObj.ID',FSavedObj.ID);
    {$ENDIF}
    AddScrollLine(FSavedObj.Name);
  end;
end;

procedure TScrollBarPlugIn.OnMouseWheel(zDelta: Integer);
begin

  if not FScrolling and FEnabled then
  begin
 {$IFDEF Codesite}
   Codesite.SendInteger('MSWheel',zDelta);
 {$ENDIF}
    FScrolling := True;
//    FHSaveScrollPOS := GetScrollPos(FSavedObj.ID,SB_HORZ);
//    FVSaveScrollPOS := GetScrollPos(FSavedObj.ID,SB_VERT);
  end;
end;

procedure TScrollBarPlugIn.OnRecordPause;
begin
  FEnabled := False;
end;

procedure TScrollBarPlugIn.OnRecordResume;
begin
  FEnabled := True;
  FSavedObj  := PIManager.SYS.ActiveWindow;
end;

procedure TScrollBarPlugIn.OnRecordStart;
begin
{$IFDEF Codesite}
CodeSite.EnterMethod(Self, 'OnRecordStart');
{$ENDIF}
  FScriptLang := PIManager.Project.ScriptLang;
  FEnabled := True;
  if @FGetHookRec <> nil then
  begin
    FHookRec := FGetHookRec;
    if FHookRec <> nil then
    begin
      FHookRec^.TheCallBack := FHWND;
    end
    else
      MessageBox(0, 'Unable to Assign Callback', 'ScrollBars', MB_OK + MB_ICONERROR);
  end;
  if @FStartHook <> nil then
    FStartHook;
  FSavedObj  := PIManager.SYS.ActiveWindow;
{$IFDEF Codesite}
CodeSite.ExitMethod(Self, 'OnRecordStart');
{$ENDIF}
end;

procedure TScrollBarPlugIn.OnRecordStop;
begin
{$IFDEF Codesite}
 CodeSite.SendMsg('OnRecordStop');
{$ENDIF}
  if FScrolling and FEnabled then
    AddScrollLine(FSavedObj.Name);
  FEnabled := False;
  if @FStopHooK <> nil then
    FStopHook;
end;



procedure TScrollBarPlugIn.UpdateSettings;
begin

end;

procedure TScrollBarPlugIn.WinProc(var Message: TMessage);
var
  MyWin:OLEVariant;
  MyMouse:TWMMOuse;
begin
  if Message.Msg = WM_CALL then
  begin
    FScrolling := FEnabled;
  end;
end;


{ export functions }

function AQtestPlugInInit(PlugInManager: IPIBaseManager): WordBool;
begin
  Result := True;
  try
    PIManager := PlugInManager as IPIManager;
    ScrollBarPlugIn := TScrollBarPlugIn.Create;
    PIManager.Engines.AddEngine(SCROLLBARPlugIn);
    PIManager.AddName('ScrollBars', True, FSCROLLBARS);
    PIManager.AddNotifier(IID_IPTEngineNotifier, SCROLLBARPlugIn);
    PIManager.AddNotifier(IID_IPTLowRecorderNotifier,SCROLLBARPlugIn);
  except
    Result := False;
  end;
end;

procedure AQtestPlugInFinalize;
begin
  PIManager.RemoveNotifier(IID_IPTLowRecorderNotifier,SCROLLBARPlugIn);
  PIManager.RemoveNotifier(IID_IPTEngineNotifier, SCROLLBARPlugIn);
  PIManager.Engines.RemoveEngine(SCROLLBARPlugIn);
  PIManager.RemoveName('ScrollBars');
//  ScrollBarPlugIn.Free;
//  ScrollBarPlugIn := nil;
end;

function AQtestPlugInDescription: WideString;
begin
  Result := 'Scrollbars Support'
end;





end.
